

/**
 * cudaSyM40
 * Martin Mattsson
 * This sample is a very basic sample that implements element by element
 * vector addition. It is the same as the sample illustrating Chapter 2
 * of the programming guide with some additions like error checking.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
// For the CUDA runtime routines (prefixed with "cuda_")
// includes, project
#include <helper_functions.h>
#include <helper_cuda.h>

#include <cuda_runtime.h>

//
#include <cuda_runtime_api.h>
#include <cooperative_groups.h>
#include "SyMConn10Lib.h"

#include <iostream>
#include <fstream>
using namespace std;
namespace cg = cooperative_groups;
const char *sSDKname = "SyM40CG_connectivity";
/**
 * CUDA Kernel Device code
 *__device__ functions can be called only from the device, and it is executed only in the device.
 *__global__ functions can be called from the host, and it is executed in the device.
 */
//vectorAdd<<<blocksPerGrid, threadsPerBlock>>>(d_P, d_E, d_K, d_R, numElements);


__device__ void cudaVehicleLight(
	const cg::grid_group &grid,
	int nrRays,
	int d_xy_map,
	//int d_y_map,
	int Ylength,
	int Xlength,
	//signed int *d_isoC,
	//unsigned short *d_CM,
	double *d_isoC,	//Isocandela UDdeg, Rldeg, Lum_cd, x_cell, y_cell, value
	double *d_LmHO,	//LM elvation map + objects
	double d_LMlocalH, // vehicle elevation
	double d_Tintensity, //Traffic intensity, vehicles per second in 10mx10m
	double d_RoadDirection, //road direction deg
	double d_Roadslope, // roadslope deg
	double d_sumLuminance, // sum luminance from 1 vehicle
	double d_watt // light watt per vehicle
	//double *d_R
)
{

	

	for (int i = grid.thread_rank(); i < nrRays; i += grid.size()) {
		
		int d_x_local;
		int d_y_local;
		int d_xy_receiver;
		double d_z_local;
		double d_UDrad; //vertical raylight part
		double d_RLrad;	//horisontal raylight part
		double d_LmH_local;
		double PI = 3.1415926536;

		d_UDrad = (d_Roadslope + d_isoC[i]) * (PI / 180);
		d_RLrad = (d_RoadDirection + d_isoC[i + d_nrRays]) * (PI / 180);

		d_isoC[i + (4 * d_nrRays)] = 0;
		d_isoC[i + (5 * d_nrRays)] = 0;
		d_isoC[i + (6 * d_nrRays)] = 0;

		for (int j = 0; j < 200; j++) {

			// xcols= (COS(RADIANER(B2)))*A2 * 0, 1
			// ycols=(SIN(RADIANER(B2)))*A2*0,1

			// horisontal rayligth
			d_x_local = floor(cos(d_RLrad) * j * 0.1); // 0.1 10mx10m grid
			d_y_local = floor(sin(d_RLrad) * j * 0.1);

			//vertical raylight
			
			// control if xy local is within map
			d_xy_receiver = d_xy_map + (d_x_local);
			d_z_local = d_LmHO[d_xy_receiver] + (sin(d_UDrad) * j);

			if ((d_z_local) >= (d_LMlocalH + 0.5)) {
				j = 200;
				d_isoC[i + (4 * d_nrRays)] = d_x_local;
				d_isoC[i + (5 * d_nrRays)] = d_y_local;
				d_isoC[i + (6 * d_nrRays)] = (d_isoC[i + (3 * d_nrRays)] / d_sumLuminance) * d_Tintensity * d_watt;
			}
		}
		
	}
}

extern "C" __global__ void cudaWiChVehicleLight(
	int nrRays,
	int d_xy_map,
	//int d_y_map,
	int Ylength,
	int Xlength,
	//signed int *d_isoC,
	//unsigned short *d_CM,
	double *d_isoC,	//Isocandela UDdeg, Rldeg, Lum_cd, x_cell, y_cell, value
	double *d_LmHO,	//LM elvation map + objects
	double d_LMlocalH, // vehicle elevation
	double d_Tintensity, //Traffic intensity, vehicles per second in 10mx10m
	double d_RoadDirection, //road direction deg
	double d_Roadslope, // roadslope deg
	double d_sumLuminance, // sum luminance from 1 vehicle
	double d_watt // light watt per vehicle
	//double *d_R
)
{
    // Handle to thread block group
	cg::thread_block block = cg::this_thread_block();
	cg::grid_group grid = cg::this_grid();

	cudaVehicleLight(
	grid,
	nrRays,
	d_xy_map,
	//int d_y_map,
	Ylength,
	Xlength,
	//signed int *d_isoC,
	//unsigned short *d_CM,
	d_isoC,	//Isocandela UDdeg, Rldeg, Lum_cd, x_cell, y_cell, value
	d_LmHO,	//LM elvation map + objects
	d_LMlocalH, // vehicle elevation
	d_Tintensity, //Traffic intensity, vehicles per second in 10mx10m
	d_RoadDirection, //road direction deg
	d_Roadslope, // roadslope deg
	d_sumLuminance, // sum luminance from 1 vehicle
	d_watt // light watt per vehicle
	//d_R
);

cg::sync(grid);



}
////////////////////////////////////////////////////////////////////////////////
// Wrapper function for kernel launch
////////////////////////////////////////////////////////////////////////////////
//call_cudaSyM40(numElements, numThreads, numBlocks, Plength, Elength, d_P, d_E, d_K, d_R, d_IR, d_Rtemp, d_M, d_F, d_V);
void call_cudaWiChVehicleLight(
	int nrRays,
	int d_xy_map,
	//int d_y_map,
	int Ylength,
	int Xlength,
	//signed int *d_isoC,
	//unsigned short *d_CM,
	double *d_isoC,	//Isocandela UDdeg, Rldeg, Lum_cd, x_cell, y_cell, value
	double *d_LmHO,	//LM elvation map + objects
	double d_LMlocalH, // vehicle elevation
	double d_Tintensity, //Traffic intensity, vehicles per second in 10mx10m
	double d_RoadDirection, //road direction deg
	double d_Roadslope, // roadslope deg
	double d_sumLuminance, // sum luminance from 1 vehicle
	double d_watt // light watt per vehicle
	//double *d_R
)

{
    int smemSize = threads * sizeof(double);
	void *kernelArgs[] = {
		(void*)&nrRays,
		(void*)&d_xy_map,
		(void*)&Ylength,
		(void*)&Xlength,
		(void*)&d_isoC,
		(void*)&d_LmHO,
		(void*)&d_LMlocalH,
		(void*)&d_Tintensity,
		(void*)&d_RoadDirection,
		(void*)&d_Roadslope,
		(void*)&d_sumLuminance,
		(void*)&d_watt
		//(void*)&d_R,
	};

    dim3 dimBlock(threads, 1, 1);
    dim3 dimGrid(numBlocks, 1, 1);

    cudaLaunchCooperativeKernel((void*)cudaWiChVehicleLight, dimGrid, dimBlock, kernelArgs, smemSize, NULL);
}

unsigned int nextPow2(unsigned int x)
{
    --x;
    x |= x >> 1;
    x |= x >> 2;
    x |= x >> 4;
    x |= x >> 8;
    x |= x >> 16;
    return ++x;
}

////////////////////////////////////////////////////////////////////////////////
// Compute the number of threads and blocks to use for the reduction
// We set threads / block to the minimum of maxThreads and n/2.
////////////////////////////////////////////////////////////////////////////////
void getNumBlocksAndThreads(int n, int maxBlocks, int maxThreads, int &blocks, int &threads)
{
    if (n == 1)
    {
        threads = 1;
        blocks = 1;
    }
    else
    {
        threads = (n < maxThreads*2) ? nextPow2(n / 2) : maxThreads;
        blocks = max(1, n / (threads * 2));
    }

    blocks = min(maxBlocks, blocks);
}

void WiChVehicleLight(
	int nrRays,
	//int d_xy_map,
	//int d_y_map,
	int Ylength,
	int Xlength,
	//signed int *d_isoC,
	//unsigned short *d_CM,
	double *h_isoC,	//Isocandela UDdeg, Rldeg, Lum_cd, x_cell, y_cell, value
	double *h_LmHO,	//LM elvation map + objects
	double *h_LMlocalH, // vehicle elevation
	double *h_Tintensity, //Traffic intensity, vehicles per second in 10mx10m
	double *h_RoadDirection, //road direction deg
	double *h_Roadslope, // roadslope deg
	double sumLuminance, // sum luminance from 1 vehicle
	double watt, // light watt per vehicle
	double *h_Rmap, //result
	int nzz)
{
	int device;
	cudaDeviceProp deviceProp = { 0 };
	//printf("%s Starting...\n\n", sSDKname);

	ofstream errfile;
	errfile.open("err20190520.txt", ofstream::app);
	errfile << "error\n";
	errfile.close();

	device = 0;
	checkCudaErrors(cudaGetDeviceProperties(&deviceProp, device));
	if (!deviceProp.cooperativeLaunch)
	{
		//printf("\nSelected GPU (%d) does not support Cooperative Kernel Launch, Waiving the run\n", device);
		exit(EXIT_WAIVED);
	}
	cudaError_t err = cudaSuccess;
	// vector length to be used, and compute its size
	int numElements = nrRays;
	size_t isoCsize = nrRays * sizeof(double);

	size_t LmHOsize = Ylength * Xlength * sizeof(double);

	//size_t Rmapsize = Ylength * Xlength * sizeof(double);
	//size_t cooDsize = coolength * sizeof(float);
	//size_t LRsize = 3 * coolength * sizeof(int);
	//int *h_LR = LRsize;
	//int *h_LR = (int *)malloc(LRsize);
	
	//for (int i = 0; i < coolength * 3; ++i)
	//{
	//	h_LR[i] = h_fLR[i];

	//}

	// Allocate the device input/output vector
	double *d_isoC = NULL;
	err = cudaMalloc((void **)&d_isoC, isoCsize);
	double *d_LmHO = NULL;
	err = cudaMalloc((void **)&h_LmHO, LmHOsize);


	//unsigned short *d_CM = NULL;
	//err = cudaMalloc((void **)&d_CM, SGsize);
	//float *d_Rso = NULL;
	//err = cudaMalloc((void **)&d_Rso, cooDsize);
	//float *d_Rsi = NULL;
	//err = cudaMalloc((void **)&d_Rsi, cooDsize);
	//float *d_ISm = NULL;
	//err = cudaMalloc((void **)&d_ISm, cooDsize);
	//signed int *d_LR = NULL;
	//err = cudaMalloc((void **)&d_LR, LRsize);

	//float *d_RSGso = NULL;
	//err = cudaMalloc((void **)&d_RSGso, RSGsize);
	//float *d_RSGsi = NULL;
	//err = cudaMalloc((void **)&d_RSGsi, RSGsize);

	// Copy the host input vectors in host memory to the device input vectors in device memory
	err = cudaMemcpy(d_isoC, h_isoC, isoCsize, cudaMemcpyHostToDevice);
	err = cudaMemcpy(d_LmHO, h_LmHO, LmHOsize, cudaMemcpyHostToDevice);

	// Determine the launch configuration (threads, blocks)
	int maxThreads = 0;
	int maxBlocks = 0;

	maxThreads = deviceProp.maxThreadsPerBlock;
	maxBlocks = deviceProp.multiProcessorCount * (deviceProp.maxThreadsPerMultiProcessor / deviceProp.maxThreadsPerBlock);

	int numBlocks = 0;
	int numThreads = 0;
	getNumBlocksAndThreads(numElements, maxBlocks, maxThreads, numBlocks, numThreads);

	// We calculate the occupancy to know how many block can actually fit on the GPU
	int numBlocksPerSm = 0;
	err = cudaOccupancyMaxActiveBlocksPerMultiprocessor(&numBlocksPerSm, cudaWiChVehicleLight, numThreads, numThreads * sizeof(double));


	int numSms = deviceProp.multiProcessorCount;
	if (numBlocks > numBlocksPerSm * numSms)
	{
		numBlocks = numBlocksPerSm * numSms;
	}

	double i_h_LMlocalH = 0;
	double i_h_Tintensity = 0;
	double i_h_RoadDirection = 0;
	double i_h_Roadslope = 0;
	int tx;
	int ty;
	//-----------------------------------------------------------------------------------
	// Gör denna xy baserad lättare att ta bort outliners
	for (int i = 0; i < nzz; ++i) {
	
		if (h_Tintensity > 0) {
			i_h_LMlocalH = h_LMlocalH[i];
			i_h_Tintensity = h_Tintensity[i];
			i_h_RoadDirection = h_RoadDirection[i];
			i_h_Roadslope = h_Roadslope[i];

			call_cudaWiChVehicleLight(
				nrRays,
				//numElements,
				//numThreads,
				//numBlocks,
				i, //int d_xy_map,
				//int d_y_map,
				Ylength,
				Xlength,
				//signed int *d_isoC,
				//unsigned short *d_CM,
				d_isoC,	//Isocandela UDdeg, Rldeg, Lum_cd, x_cell, y_cell, value
				d_LmHO,	//LM elvation map + objects
				i_h_LMlocalH, // vehicle elevation
				i_h_Tintensity, //Traffic intensity, vehicles per second in 10mx10m
				i_h_RoadDirection, //road direction deg
				i_h_Roadslope, // roadslope deg
				sumLuminance, // sum luminance from 1 vehicle
				watt // light watt per vehicle
			);
			cudaDeviceSynchronize();

			err = cudaMemcpy(h_isoC, d_isoC, isoCsize, cudaMemcpyDeviceToHost);
			for (int j = 0; j < nrRays; j++) {

				//h_Rmap[i] = h_Rmap[i] + h_isoC[j];
				// light data in 6
				tx = int(h_isoC[j + (4 * nrRays)]);
				ty = int(h_isoC[j + (5 * nrRays)]);

				h_Rmap[i] = h_Rmap[i] + h_isoC[j + (6 * nrRays)];

			}

		}

		
	}

	//float *h_cooD,
	//	int *h_cooR,
	//	int *h_cooC, h_Rso, h_Rsi
	//M * v
	//row, col, val
	//sannolikhet att spridas från rad i till kolumn j
	//OBS Sparse matrise is from matlab with start 1 (not zero based)
	// result[Row[k]-1] = result[Row[k]-1] + Val[k]*d[Col[k]-1];
	//12.08 medel area km2 0.250*0.250=0.625km2 0.625/12.08 =0.0051738

	for (int i = 1; i < nzz; i++) {

		//h_Rso[h_cooC[i]-1] = h_Rso[h_cooC[i]-1] + h_cooD[i] * h_ISm[h_cooR[i]-1];

		//h_Rsi[h_cooR[i]-1] = h_Rsi[h_cooR[i]-1] + h_cooD[i] * h_ISm[h_cooC[i]-1];
		h_Rsi[h_cooC[i]-1] = h_Rsi[h_cooC[i]-1] + h_cooD[i] * h_ISm[h_cooR[i]-1];

		h_Rso[h_cooR[i]-1] = h_Rso[h_cooR[i]-1] + h_cooD[i] * h_ISm[h_cooC[i]-1];

	}
	err = cudaMemcpy(d_Rso, h_Rso, cooDsize, cudaMemcpyHostToDevice);
	err = cudaMemcpy(d_Rsi, h_Rsi, cooDsize, cudaMemcpyHostToDevice);
	errfile.open("err20190520.txt", ofstream::app);
	errfile << "d\n";
	errfile.close();
	//-----------------------------------------------------------------------------------
	eqnr = 2;
	//call_cudaSyM40(numElements, numThreads, numBlocks, Plength, Elength, d_P, d_E, d_K, d_R, d_IR, d_Rtemp, d_M, d_F, d_V, maxValue, d_R100);
	call_SyMConn10(
		eqnr,
		numElements,
		numThreads,
		numBlocks,
		coolength,
		Ylength,
		Xlength,
		d_LR,
		d_CM,
		d_IS,
		d_ISm,
		d_Rso,
		d_Rsi,
		d_RSGso,
		d_RSGsi);

	cudaDeviceSynchronize();
	errfile.open("err20190520.txt", ofstream::app);
	errfile << "f\n";
	errfile.close();
	//-----------------------------------------------------------------------------------
	// Copy the device result vector in device memory to the host result vector in host memory.
	err = cudaMemcpy(h_Rso, d_Rso, ISsize, cudaMemcpyDeviceToHost);
	err = cudaMemcpy(h_Rsi, d_Rsi, ISsize, cudaMemcpyDeviceToHost);

	err = cudaMemcpy(h_RSGso, d_RSGso, ISsize, cudaMemcpyDeviceToHost);
	err = cudaMemcpy(h_RSGsi, d_RSGsi, ISsize, cudaMemcpyDeviceToHost);

	//err = cudaMemcpy(h_R100, d_R100, Rsize, cudaMemcpyDeviceToHost);
	errfile.open("err20190520.txt", ofstream::app);
	errfile << "\n";
	for (int i = 0; i < coolength; ++i)
	{

		//errfile << h_Rso[i];
		//errfile << "\n";

	}
	errfile.close();
	// Free device global memory
	err = cudaFree(d_LR);
	err = cudaFree(d_CM);
	err = cudaFree(d_IS);
	err = cudaFree(d_ISm);
	err = cudaFree(d_Rso);
	err = cudaFree(d_Rsi);
	err = cudaFree(d_RSGso);
	err = cudaFree(d_RSGsi);


	// Free host memory
	free(h_LR);
	//free(h_CM);
	//free(h_IS);
	//free(h_ISm);
	//free(h_Rso);
	//free(h_Rsi);


}

